# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Interface for collections."""

import datetime as dt
import re
from abc import ABC
from collections.abc import Iterable
from typing import Any, override

from django.db.models import Q

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    BaseArtifactDataModel,
    CollectionCategory,
)
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    User,
    WorkRequest,
    Workspace,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.tasks.models import LookupMultiple, LookupSingle


class ItemAdditionError(Exception):
    """
    Raised if the object cannot be added to the collection.

    For example: the object's category is not valid for the collection,
    or adding this object breaks some collection constraints.
    """


class ItemRemovalError(Exception):
    """
    Raised if the object cannot be removed from the collection.

    For example: removing this Item breaks a collection constraint.
    """


class CollectionManagerInterface(ABC):
    """Interface to manage a collection."""

    # Category of the managed collection
    COLLECTION_CATEGORY: CollectionCategory

    # Valid categories of the bare data items added in the collection; if
    # None, any category is valid
    VALID_BARE_DATA_CATEGORIES: frozenset[BareDataCategory] | None = frozenset()

    # Valid categories of the artifacts added in the collection; if None,
    # any category is valid
    VALID_ARTIFACT_CATEGORIES: frozenset[ArtifactCategory] | None = frozenset()

    # Valid categories of the collections added in the collection; if None,
    # any category is valid
    VALID_COLLECTION_CATEGORIES: frozenset[CollectionCategory] | None = (
        frozenset()
    )

    MANAGERS: dict[str, type["CollectionManagerInterface"]] = {}

    def __init__(self, collection: Collection) -> None:
        """
        Instantiate collection manager.

        :param collection: collection to be managed
        """
        if collection.category != self.COLLECTION_CATEGORY:
            raise ValueError(
                f'{self.__class__.__name__} cannot manage '
                f'"{collection.category}" category'
            )

        # collection that the items will be added / removed / looked up
        self.collection = collection

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Register the subclass."""
        super().__init_subclass__(**kwargs)
        cls.MANAGERS[cls.COLLECTION_CATEGORY] = cls

    @override
    def __eq__(self, other: Any) -> bool:
        """Two managers are equal if they manage the same collection."""
        if not isinstance(other, self.__class__):
            return False
        return self.collection == other.collection

    @classmethod
    def get_manager_for(
        cls, collection: Collection
    ) -> "CollectionManagerInterface":
        """Get specialized manager subclass."""
        manager_cls = cls.MANAGERS[collection.category]
        return manager_cls(collection)

    def add_bare_data(
        self,
        category: BareDataCategory,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        data: BaseArtifactDataModel | dict[str, Any] | None = None,
        name: str | None = None,
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        """
        Add bare data to the managed collection.

        Verify that the item can be added (e.g. bare data category is valid)
        and call
        `self.do_add_bare_data(user=user, variables=variables)`.

        :param category: category of the bare data
        :param user: user adding the data to the collection
        :param workflow: workflow adding the data to the collection
        :param data: per-item data; the specific collection may also use
          this to compute the item name
        :param name: set name of new collection item (implementations may
          ignore this if they compute their own names)
        :param replace: if True, replace an existing similar item
        :param created_at: if set, mark the new item as having been created
          at this timestamp rather than now
        :param replaced_by: if set, mark the new item as having been
          replaced by this more-recently-created item
        """
        if (
            self.VALID_BARE_DATA_CATEGORIES is not None
            and category not in self.VALID_BARE_DATA_CATEGORIES
        ):
            raise ItemAdditionError(
                f'Bare data category "{category}" '
                f'not supported by the collection'
            )
        if isinstance(data, BaseArtifactDataModel):
            data = data.model_dump(mode="json", exclude_unset=True)

        return self.do_add_bare_data(
            category,
            user=user,
            workflow=workflow,
            data=data,
            name=name,
            replace=replace,
            created_at=created_at,
            replaced_by=replaced_by,
        )

    def add_artifact(
        self,
        artifact: Artifact,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        variables: dict[str, Any] | None = None,
        name: str | None = None,
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        """
        Add the artifact to the managed collection.

        Verify that the artifact can be added (e.g. artifact category is valid)
        and call
        `self.do_add_artifact(artifact, user=user, variables=variables)`.

        :param artifact: artifact to add
        :param user: user adding the artifact to the collection
        :param workflow: workflow adding the artifact to the collection
        :param variables: additional variables passed down to the specific
          collection's :py:meth:`do_add_artifact`, which may be used to
          compute per-item data
        :param name: set name of new collection item (implementations may
          ignore this if they compute their own names)
        :param replace: if True, replace an existing similar item
        :param created_at: if set, mark the new item as having been created
          at this timestamp rather than now
        :param replaced_by: if set, mark the new item as having been
          replaced by this more-recently-created item
        """
        if (
            self.VALID_ARTIFACT_CATEGORIES is not None
            and artifact.category not in self.VALID_ARTIFACT_CATEGORIES
        ):
            raise ItemAdditionError(
                f'Artifact category "{artifact.category}" '
                f'not supported by the collection'
            )

        return self.do_add_artifact(
            artifact,
            user=user,
            workflow=workflow,
            variables=variables,
            name=name,
            replace=replace,
            created_at=created_at,
            replaced_by=replaced_by,
        )

    def add_collection(
        self,
        collection: Collection,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        variables: dict[str, Any] | None = None,
        name: str | None = None,
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        """
        Add the collection to the managed collection.

        Verify that the collection can be added (e.g. collection category
        is valid) and call
        `self.do_add_collection(collection, user=user, variables=variables)`.

        :param collection: collection to add
        :param user: user adding the collection
        :param workflow: workflow adding the collection to the collection
        :param variables: additional variables passed down to the specific
          collection's :py:meth:`do_add_collection`, which may be used to
          compute per-item data
        :param name: set name of new collection item (implementations may
          ignore this if they compute their own names)
        :param replace: if True, replace an existing similar item
        :param created_at: if set, mark the new item as having been created
          at this timestamp rather than now
        :param replaced_by: if set, mark the new item as having been
          replaced by this more-recently-created item
        """
        if (
            self.VALID_COLLECTION_CATEGORIES is not None
            and collection.category not in self.VALID_COLLECTION_CATEGORIES
        ):
            raise ItemAdditionError(
                f'Collection category "{collection.category}" '
                f'not supported by the collection'
            )

        return self.do_add_collection(
            collection,
            user=user,
            workflow=workflow,
            variables=variables,
            name=name,
            replace=replace,
            created_at=created_at,
            replaced_by=replaced_by,
        )

    def remove_item(
        self,
        item: CollectionItem,
        *,
        user: User | None = None,
        workflow: WorkRequest | None = None,
    ) -> None:
        """
        Remove an item from the managed collection.

        Verify that the item can be removed and call
        self.do_remove_item(user=user).

        :param item: item to remove from the collection
        :param user: user removing the item from the collection
        :param workflow: workflow removing the item from the collection
        """
        # Check that the item can be removed
        # raise ItemRemovalError() if needed
        self.do_remove_item(item, user=user, workflow=workflow)

    def remove_items_by_name(
        self,
        name: str,
        child_types: Iterable[CollectionItem.Types],
        *,
        user: User | None = None,
        workflow: WorkRequest | None = None,
    ) -> None:
        """Remove items from the collection by name."""
        for item in self.collection.child_items.active().filter(
            child_type__in=child_types, name=name
        ):
            self.remove_item(item, user=user, workflow=workflow)

    def lookup(self, query: str) -> CollectionItem | None:
        """
        Return one CollectionItem.

        For example, a caller could do::

          collection_manager.lookup(
              "match:codename=bookworm:architecture=amd64:variant=apt"
          )

        :param query: `name:NAME` to find an active item with the given
          name.  Specific collection implementations may define other valid
          queries.
        :return: a single CollectionItem, or None if the query is in a
          recognized format but does not match an item.
        :raise LookupError: the query contains an unknown lookup type.
        """
        if m := re.match(r"^name:(.+)$", query):
            try:
                return self.collection.child_items.get(
                    removed_at__isnull=True, name=m.group(1)
                )
            except CollectionItem.DoesNotExist:
                return None
        return self.do_lookup(query)

    def lookup_filter(
        self,
        key: str,
        value: LookupSingle | LookupMultiple,
        *,
        workspace: Workspace,
        pc: PermissionContext,
        workflow_root: WorkRequest | None = None,
    ) -> Q:
        """
        Return :py:class:`CollectionItem` conditions for a lookup filter.

        :param key: name of the lookup filter
        :param value: parameter passed to the specific lookup filter
        :param workspace: workspace to perform the lookup in
        :param user: user performing the lookup
        :param workflow_root: root of the workflow performing the lookup, if
          any
        :return: a Django :py:class:`Q` object with conditions for this lookup
          filter.
        :raise LookupError: the query contains an unknown lookup filter type.
        """
        return self.do_lookup_filter(
            key,
            value,
            workspace=workspace,
            pc=pc,
            workflow_root=workflow_root,
        )

    def do_add_bare_data(
        self,
        category: BareDataCategory,  # noqa: ARG002, U100
        *,
        user: User,  # noqa: ARG002, U100
        workflow: WorkRequest | None,  # noqa: ARG002, U100
        data: dict[str, Any] | None = None,  # noqa: ARG002, U100
        name: str | None = None,  # noqa: ARG002, U100
        replace: bool = False,  # noqa: ARG002, U100
        created_at: dt.datetime | None = None,  # noqa: ARG002, U100
        replaced_by: CollectionItem | None = None,  # noqa: ARG002, U100
    ) -> CollectionItem:  # pragma: no cover
        """
        Add bare data into the managed collection.

        Called by add_bare_data().

        :raise ItemAdditionError: the item cannot be added: breaks
          collection's constraints.
        """
        raise ItemAdditionError(
            f'Cannot add bare data into "{self.COLLECTION_CATEGORY}"'
        )

    def do_add_artifact(
        self,
        artifact: Artifact,  # noqa: ARG002, U100
        *,
        user: User,  # noqa: ARG002, U100
        workflow: WorkRequest | None = None,  # noqa: ARG002, U100
        variables: dict[str, Any] | None = None,  # noqa: ARG002, U100
        name: str | None = None,  # noqa: ARG002, U100
        replace: bool = False,  # noqa: ARG002, U100
        created_at: dt.datetime | None = None,  # noqa: ARG002, U100
        replaced_by: CollectionItem | None = None,  # noqa: ARG002, U100
    ) -> CollectionItem:  # pragma: no cover
        """
        Add the artifact into the managed collection.

        Called by add_artifact().

        :raise ItemAdditionError: the artifact cannot be added: breaks
          collection's constraints.
        """
        raise ItemAdditionError(
            f'Cannot add artifacts into "{self.COLLECTION_CATEGORY}"'
        )

    def do_add_collection(
        self,
        collection: Collection,  # noqa: ARG002, U100
        *,
        user: User,  # noqa: ARG002, U100
        workflow: WorkRequest | None = None,  # noqa: ARG002, U100
        variables: dict[str, Any] | None = None,  # noqa: ARG002, U100
        name: str | None = None,  # noqa: ARG002, U100
        replace: bool = False,  # noqa: ARG002, U100
        created_at: dt.datetime | None = None,  # noqa: ARG002, U100
        replaced_by: CollectionItem | None = None,  # noqa: ARG002, U100
    ) -> CollectionItem:
        """
        Add the collection into the managed collection.

        Called by add_collection().

        :raise ItemAdditionError: the collection cannot be added: breaks
          collection's constraints.
        """
        raise ItemAdditionError(
            f'Cannot add collections into "{self.COLLECTION_CATEGORY}"'
        )

    def do_remove_item(
        self,
        item: CollectionItem,  # noqa: ARG002, U100
        *,
        user: User | None = None,  # noqa: ARG002, U100
        workflow: WorkRequest | None = None,  # noqa: ARG002, U100
    ) -> None:  # pragma: no cover
        """
        Remove an item from the managed collection.

        Called by remove_item().

        :raise: ItemRemovalError: the item cannot be removed: breaks
          collection's constraints.
        """
        raise ItemRemovalError(
            f'Cannot remove bare data from "{self.COLLECTION_CATEGORY}"'
        )

    def do_lookup(self, query: str) -> CollectionItem | None:
        """
        Return one CollectionItem.

        Called by lookup().

        :raise LookupError: the query contains an unknown lookup type.
        """
        raise LookupError(f'Unexpected lookup format: "{query}"')

    def do_lookup_filter(
        self,
        key: str,
        value: LookupSingle | LookupMultiple,  # noqa: ARG002, U100
        *,
        workspace: Workspace,  # noqa: ARG002, U100
        pc: PermissionContext,  # noqa: ARG002, U100
        workflow_root: WorkRequest | None = None,  # noqa: ARG002, U100
    ) -> Q:
        """
        Return :py:class:`CollectionItem` conditions for a lookup filter.

        Called by lookup_filter().

        :raise LookupError: the query contains an unknown lookup filter type.
        """
        raise LookupError(f'Unexpected lookup filter format: "{key}"')



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the artifact playground functions."""

from debusine.artifacts.local_artifact import Upload
from debusine.artifacts.playground import ArtifactPlayground
from debusine.test import TestCase


class ArtifactPlaygroundTests(TestCase):
    """Test artifact playground behaviour."""

    def test_source_package_defaults(self) -> None:
        """Check building a source package."""
        workdir = self.create_temporary_directory()
        src = ArtifactPlayground.create_source_package(workdir)
        self.assertEqual(
            src.files,
            {
                'hello_1.0-1.debian.tar.xz': workdir
                / "hello_1.0-1.debian.tar.xz",
                'hello_1.0-1.dsc': workdir / "hello_1.0-1.dsc",
                'hello_1.0.orig.tar.gz': workdir / "hello_1.0.orig.tar.gz",
            },
        )
        self.assert_source_artifact_equal(src.data)

    def test_source_package_native(self) -> None:
        """Check building a source package."""
        workdir = self.create_temporary_directory()
        src = ArtifactPlayground.create_source_package(workdir, version="1.0")
        self.assertEqual(
            src.files,
            {
                'hello_1.0.dsc': workdir / "hello_1.0.dsc",
                'hello_1.0.tar.gz': workdir / "hello_1.0.tar.gz",
            },
        )
        self.assert_source_artifact_equal(src.data, version="1.0")

    def test_package_build_log(self) -> None:
        """Check building a package build log."""
        workdir = self.create_temporary_directory()
        filename = "hello_1.0-1_source.build"
        file = workdir / filename
        file.touch()
        buildlog = ArtifactPlayground.create_package_build_log(file)
        self.assertEqual(buildlog.data.source, "hello")
        self.assertEqual(buildlog.data.version, "1.0-1")
        self.assertEqual(buildlog.data.filename, filename)

    def test_package_build_log_args(self) -> None:
        """Check building a package build log overriding source/version."""
        workdir = self.create_temporary_directory()
        filename = "hello_1.0-1_source.build"
        file = workdir / filename
        file.touch()
        buildlog = ArtifactPlayground.create_package_build_log(
            file, source="test"
        )
        self.assertEqual(buildlog.data.source, "test")
        self.assertEqual(buildlog.data.version, "1.0-1")
        self.assertEqual(buildlog.data.filename, filename)

        buildlog = ArtifactPlayground.create_package_build_log(
            file, version="2.0"
        )
        self.assertEqual(buildlog.data.source, "hello")
        self.assertEqual(buildlog.data.version, "2.0")
        self.assertEqual(buildlog.data.filename, filename)

        buildlog = ArtifactPlayground.create_package_build_log(
            file, source="test", version="2.0"
        )
        self.assertEqual(buildlog.data.source, "test")
        self.assertEqual(buildlog.data.version, "2.0")
        self.assertEqual(buildlog.data.filename, filename)

    def test_package_build_log_invalid_filename(self) -> None:
        """Check creating a build log with an invalid name."""
        workdir = self.create_temporary_directory()
        (file := workdir / "foo.build").touch()
        with self.assertRaisesRegex(ValueError, "cannot be inferred"):
            ArtifactPlayground.create_package_build_log(file)

    def test_binary_package(self) -> None:
        """Check building a binary package."""
        workdir = self.create_temporary_directory()
        binary = ArtifactPlayground.create_binary_package(workdir)
        self.assertEqual(binary.data.srcpkg_name, "hello")
        self.assertEqual(binary.data.srcpkg_version, "1.0-1")
        self.assertEqual(
            binary.data.deb_fields,
            {
                'Architecture': 'all',
                'Description': 'Example description',
                'Maintainer': 'Example Maintainer <example@example.org>',
                'Package': 'hello',
                'Version': '1.0-1',
            },
        )
        self.assertEqual(binary.data.deb_control_files, ["control"])

    def assertUpload(
        self,
        upload: Upload,
        *,
        expected_version: str = "1.0-1",
        expected_architecture: set[str] = {"source"},
        expected_files: set[str],
        expected_binaries: set[str] = set(),
    ) -> None:
        """Assert that upload has the expected data and files."""
        data_fields = upload.data.changes_fields.copy()
        files = data_fields.pop('Files')
        for field in list(data_fields.keys()):
            if field.startswith('Checksums-'):
                data_fields.pop(field)
        expected_fields = {
            "Format": "1.8",
            "Date": "Sun, 01 Dec 2024 00:00:00 -0000",
            "Source": "hello",
            "Architecture": " ".join(sorted(expected_architecture)),
            "Version": expected_version,
            "Distribution": "unstable",
            "Urgency": "medium",
            "Maintainer": "Example Maintainer <example@example.org>",
            "Changes": "\n".join(
                (
                    "",
                    f"  hello ({expected_version}) unstable; urgency=medium",
                    "  .",
                    "    * Test upload.",
                )
            ),
        }
        if expected_binaries:
            expected_fields["Binary"] = " ".join(sorted(expected_binaries))
            expected_fields["Description"] = "\n" + "\n".join(
                f" {binary} - A Description" for binary in expected_binaries
            )
        self.assertEqual(
            data_fields,
            expected_fields,
        )
        file_names = {file["name"] for file in files}
        self.assertEqual(file_names, expected_files)

    def test_source_upload(self) -> None:
        """Check building a source-only upload."""
        workdir = self.create_temporary_directory()
        upload = ArtifactPlayground.create_upload(workdir)
        self.assertUpload(
            upload,
            expected_files={
                "hello_1.0-1.dsc",
                "hello_1.0-1.debian.tar.xz",
                "hello_1.0.orig.tar.gz",
            },
        )

    def test_source_native_upload(self) -> None:
        """Check building a source-only native upload."""
        workdir = self.create_temporary_directory()
        upload = ArtifactPlayground.create_upload(workdir, version="1.0")
        self.assertUpload(
            upload,
            expected_version="1.0",
            expected_files={
                "hello_1.0.dsc",
                "hello_1.0.tar.gz",
            },
        )

    def test_binary_upload(self) -> None:
        """Check building a binary-only upload."""
        workdir = self.create_temporary_directory()
        upload = ArtifactPlayground.create_upload(
            workdir,
            source=False,
            binary=True,
        )
        self.assertUpload(
            upload,
            expected_architecture={"all"},
            expected_files={
                "hello_1.0-1_all.deb",
            },
            expected_binaries={"hello"},
        )

    def test_mixed_upload(self) -> None:
        """Check building a source+binary upload."""
        workdir = self.create_temporary_directory()
        upload = ArtifactPlayground.create_upload(
            workdir, binary=True, binaries=[("hello", "all")]
        )
        self.assertUpload(
            upload,
            expected_architecture={"source", "all"},
            expected_files={
                "hello_1.0-1.dsc",
                "hello_1.0-1.debian.tar.xz",
                "hello_1.0.orig.tar.gz",
                "hello_1.0-1_all.deb",
            },
            expected_binaries={"hello"},
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Tests for ServerNotifications."""

import asyncio
import logging
import signal
from typing import Any, override
from unittest import mock
from unittest.mock import MagicMock

import tenacity

from debusine.client.debusine import Debusine, ServerNotificationDecodeError
from debusine.client.server_notifications import ServerNotifications, logger
from debusine.test import TestCase
from debusine.worker.config import ConfigHandler
from debusine.worker.tests import server
from debusine.worker.tests.server import ServerConfig


class MainLoopKeepsRetrying(Exception):
    """Exception raised when tenacity does not stop retrying."""


class ServerNotificationsTests(TestCase, server.DebusineAioHTTPTestCase):
    """Test :py:class:`ServerNotifications`."""

    @override
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.api_url = str(self.server.make_url('')) + "/api"
        self.temp_config_directory = self.create_temp_config_directory(
            {'General': {'api-url': self.api_url}}
        )
        self.config = ConfigHandler(
            directories=[self.temp_config_directory.as_posix()]
        )
        self.config['General']['log-file'] = str(
            self.temp_config_directory / 'worker.log'
        )

        # Set a valid token in the test server and client configuration
        token = '6c931875627131b5135b7de3371c44'
        self.server_config.registered_token = token
        self.config.write_token(token)

        self.default_sigint_handler = signal.getsignal(signal.SIGINT)
        self.default_sigterm_handler = signal.getsignal(signal.SIGTERM)

        self._server_notifications: ServerNotifications | None = None

    @override
    def tearDown(self) -> None:
        super().tearDown()
        # Restore signal handlers. Cli.execute() changes them
        signal.signal(signal.SIGINT, self.default_sigint_handler)
        signal.signal(signal.SIGTERM, self.default_sigterm_handler)

    @override
    async def tearDownAsync(self) -> None:
        if self._server_notifications:
            await self._server_notifications.close()
        await super().tearDownAsync()

    @property
    def server_notifications(self) -> ServerNotifications:
        """Return the ServerNotifications being tested."""
        if self._server_notifications is None:
            debusine_client = Debusine(
                self.config.api_url,
                self.config.token,
                timeout=self.config.http_timeout,
                logger=logging.getLogger(),
                scope=None,
            )
            self._server_notifications = ServerNotifications(
                debusine_client, endpoint="1.0/worker/connect/"
            )
            self._server_notification_asyncio_sleep = self.enterContext(
                mock.patch.object(
                    self.server_notifications,
                    "_asyncio_sleep",
                    side_effect=[None] * 5 + [MainLoopKeepsRetrying()],
                )
            )
        return self._server_notifications

    def mock_wait_for_messages(self) -> asyncio.Future[None]:
        """
        Mock the message reading function with a wait for a future.

        :returns: the future that blocks the mocked main loop: set a result to
          exit the mocked main loop, set an exception to make it raise, cancel
          it to cancel the main loop
        """
        future = asyncio.get_running_loop().create_future()

        async def mock_wait_for_messages() -> None:
            await future

        self.enterContext(
            mock.patch.object(
                self.server_notifications,
                "_wait_for_messages",
                mock_wait_for_messages,
            )
        )
        return future

    async def test_token_required(self) -> None:
        client = Debusine(
            self.config.api_url,
            None,
            timeout=self.config.http_timeout,
            logger=logging.getLogger(),
            scope=None,
        )
        with self.assertRaisesRegex(
            ValueError,
            "Debusine client needs to be configured with an API token",
        ):
            ServerNotifications(client, endpoint="1.0/worker/connect/")

    async def test_asyncio_sleep(self) -> None:
        # Reinstantiate ServerNotifications to get a version with an unpatched
        # _asyncio_sleep method
        server_notifications = ServerNotifications(
            self.server_notifications.client,
            endpoint=self.server_notifications.endpoint,
        )
        with mock.patch("asyncio.sleep") as asleep:
            await server_notifications._asyncio_sleep(4.2)
        asleep.assert_called_once_with(4.2)

    async def test_connection_success(self) -> None:
        # Do not try to reconnect when the server quits
        self.server_notifications._shutting_down = True
        await self.server_notifications.main_loop()
        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, '/api/ws/1.0/worker/connect/'
        )
        self.assertTrue(self.server_latest_request.is_authenticated)

    async def test_invalid_token(self) -> None:
        self.registered_token = 'f924ba0099588bf6'
        self.config.write_token('e3c0284')
        # Do not try to reconnect when the server quits
        self.server_notifications._shutting_down = True
        await self.server_notifications.main_loop()
        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, '/api/ws/1.0/worker/connect/'
        )
        self.assertFalse(self.server_latest_request.is_authenticated)

    async def test_connection_reset(self) -> None:
        self.enterContext(
            mock.patch.object(
                self.server_notifications,
                "_raise_try_again",
                side_effect=ConnectionResetError,
            )
        )

        log_message = 'ConnectionResetError, will reconnect'
        with (
            self.assertLogsContains(
                log_message,
                expected_count=6,
                level=logging.DEBUG,
                logger=logger,
            ),
            self.assertRaises(MainLoopKeepsRetrying),
        ):
            await self.server_notifications.main_loop()

    async def test_connection_refused(self) -> None:
        await self.server.close()

        log_message = 'Error connecting to'
        with (
            self.assertLogsContains(
                log_message, expected_count=6, logger=logger
            ),
            self.assertRaises(MainLoopKeepsRetrying),
        ):
            await self.server_notifications.main_loop()

    async def test_remote_close_exits_main_loop(self) -> None:
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES["request_dynamic_metadata"]
        ]

        async def call_close(*args: Any) -> None:
            await self.server_notifications.close()

        self.enterContext(
            mock.patch.object(
                self.server_notifications.message_queue,
                "put",
                side_effect=call_close,
            )
        )
        with (
            mock.patch.object(
                self.server_notifications, "_raise_try_again"
            ) as try_again_mock,
        ):
            await self.server_notifications.main_loop()
        try_again_mock.assert_not_called()

    async def test_token_disabled_try_again(self) -> None:
        assert self.config.token
        # The client will reconnect, so the server will keep resending this
        # sequence of messages every time
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES["token_disabled_error"],
        ]
        with (
            self.assertLogsContains(
                "The token is disabled. Will try again",
                expected_count=6,
                logger=logger,
            ),
            self.assertRaises(MainLoopKeepsRetrying),
        ):
            await self.server_notifications.main_loop()

    async def test_decode_ws_message_fails(self) -> None:
        self.enterContext(
            mock.patch.object(
                self.server_notifications.client,
                "decode_ws_message",
                side_effect=ServerNotificationDecodeError("expected error"),
            )
        )
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES["request_dynamic_metadata"],
            ServerConfig.RESPONSES["request_dynamic_metadata"],
        ]

        self.server_notifications._shutting_down = True
        with self.assertLogs(level=logging.WARNING, logger=logger) as log:
            await self.server_notifications.main_loop()

        # Messages were ignored
        self.assertEqual(self.server_notifications.message_queue.qsize(), 0)

        msg = (
            f"WARNING:{logger.name}:Unhandled websocket notification: "
            f"expected error"
        )
        self.assertEqual(log.output, [msg, msg])

    async def test_messages_enqueued(self) -> None:
        RESPONSES = ServerConfig.RESPONSES
        self.server_config.connect_response_bodies = [
            RESPONSES["connected"],
            RESPONSES["request_dynamic_metadata"],
            RESPONSES["work_request_available"],
            RESPONSES["request_dynamic_metadata"],
        ]
        # Quit the main loop when the server stops sending
        self.server_notifications._shutting_down = True
        await self.server_notifications.main_loop()

        self.assertEqual(self.server_notifications.message_queue.qsize(), 4)
        payload = self.server_notifications.message_queue.get_nowait()
        self.assertEqual(payload, RESPONSES["connected"])
        payload = self.server_notifications.message_queue.get_nowait()
        self.assertEqual(payload, RESPONSES["request_dynamic_metadata"])
        payload = self.server_notifications.message_queue.get_nowait()
        self.assertEqual(payload, RESPONSES["work_request_available"])
        payload = self.server_notifications.message_queue.get_nowait()
        self.assertEqual(payload, RESPONSES["request_dynamic_metadata"])

    def test_raise_try_again(self) -> None:
        self.assertRaises(
            tenacity.TryAgain, self.server_notifications._raise_try_again
        )

    def test_reset_retry_counter(self) -> None:
        retry_state = MagicMock()
        retry_state.attempt_number = 5

        self.server_notifications._reset_retry_attempt_number = True
        self.server_notifications._reset_retry_counter(retry_state)
        self.assertEqual(retry_state.attempt_number, 0)
        self.assertFalse(self.server_notifications._reset_retry_attempt_number)

    async def test_open(self) -> None:
        main_loop_future = self.mock_wait_for_messages()
        self.assertIsNone(self.server_notifications.main_loop_task)
        await self.server_notifications.open()
        self.assertIsNotNone(self.server_notifications.main_loop_task)
        main_loop_future.set_result(None)
        assert self.server_notifications.main_loop_task is not None
        await self.server_notifications.main_loop_task

    async def test_context_manager(self) -> None:
        self.mock_wait_for_messages()
        self.assertIsNone(self.server_notifications.main_loop_task)
        async with self.server_notifications:
            self.assertIsNotNone(self.server_notifications.main_loop_task)

    async def test_generator(self) -> None:
        RESPONSES = ServerConfig.RESPONSES
        expected = [
            RESPONSES["connected"],
            RESPONSES["request_dynamic_metadata"],
            RESPONSES["work_request_available"],
            RESPONSES["request_dynamic_metadata"],
        ]
        for payload in expected:
            await self.server_notifications.message_queue.put(payload)

        main_loop_future = self.mock_wait_for_messages()

        actual: list[dict[str, Any]] = []
        async with self.server_notifications:
            async for payload in self.server_notifications.messages():
                actual.append(payload)
                if len(actual) == len(expected):
                    main_loop_future.set_result(None)

        assert actual == expected

    async def test_close_when_iterating(self) -> None:
        RESPONSES = ServerConfig.RESPONSES
        await self.server_notifications.message_queue.put(
            RESPONSES["connected"]
        )
        await self.server_notifications.message_queue.put(
            RESPONSES["request_dynamic_metadata"]
        )

        self.mock_wait_for_messages()

        actual: list[dict[str, Any]] = []
        async with self.server_notifications:
            async for payload in self.server_notifications.messages():
                actual.append(payload)
                await self.server_notifications.close()

        assert actual == [RESPONSES["connected"]]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Sign task."""

import os
import shutil
import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, override
from unittest import mock, skipUnless

import gpg
from django.test import TestCase as DjangoTestCase
from django.test import TransactionTestCase, override_settings
from nacl.public import PrivateKey

import debusine.worker.tags as wtags
from debusine.artifacts import (
    LocalArtifact,
    SigningInputArtifact,
    SigningOutputArtifact,
    WorkRequestDebugLogs,
)
from debusine.artifacts.models import (
    ArtifactCategory,
    DebusineSigningInput,
    DebusineSigningOutput,
    EmptyArtifactData,
    SigningResult,
    WorkRequestResults,
)
from debusine.assets import AssetCategory, KeyPurpose
from debusine.client.models import (
    ArtifactResponse,
    AssetPermissionCheckResponse,
    RemoteArtifact,
)
from debusine.signing.db.models import Key
from debusine.signing.gnupg import gpg_ephemeral_context
from debusine.signing.models import SigningMode
from debusine.signing.tasks import Sign
from debusine.signing.tasks.models import SignDynamicData
from debusine.signing.tasks.tests.helpers import SigningTaskTestCase
from debusine.tasks import FALLBACK_ARCHITECTURE
from debusine.tasks.models import BaseTaskInputArtifact, WorkerType
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import create_remote_artifact
from debusine.utils import is_command_available

# Bare minimum that sbsigntool will recognize as a valid PE/COFF image.
# This is unlikely to work in any other context.
minimal_pecoff_image = (
    # DOS header:
    # DOS magic number
    b"MZ"
    # DOS header fields, ignored by sbsigntool
    + (b"\0" * 58)
    # Address of next header
    + b"\x80\0\0\0"
    # DOS message, ignored by sbsigntool
    + (b"\0" * 64)
    # PE header:
    # NT signature
    + b"PE\0\0"
    # AMD64 magic
    + b"\x64\x86"
    # Number of sections
    + b"\0\0"
    # PE header fields, ignored by sbsigntool
    + (b"\0" * 12)
    # Optional header size: 752
    + (b"\xf0\x02")
    # Flags, ignored by sbsigntool
    + b"\0\0"
    # Optional header:
    # PE32+ magic
    + b"\x0b\x02"
    # Optional header standard fields, ignored by sbsigntool
    + (b"\0" * 22)
    # Image base and section alignment, ignored by sbsigntool
    + (b"\0" * 12)
    # File alignment: 512
    + b"\0\2\0\0"
    # Optional header NT-specific fields, ignored by sbsigntool
    + (b"\0" * 20)
    # Size of headers: 1024
    + b"\0\4\0\0"
    # More optional header NT-specific fields, ignored by sbsigntool
    + (b"\0" * 48)
    # Space for five data directories (the fifth is the certificate table)
    + (b"\0" * 640)
    # Padding for alignment
    + (b"\0" * 120)
)


example_url = "http://example.org/"


class SignTests(
    SigningTaskTestCase[Sign],
    ExternalTaskHelperMixin[Sign],
    TestCase,
    DjangoTestCase,
):
    """Unit tests for the :py:class:`Sign` task."""

    SAMPLE_TASK_DATA = {
        "purpose": KeyPurpose.UEFI,
        "unsigned": [2],
        "key": "ABC123",
    }

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set_field_artifacts(
            "unsigned",
            BaseTaskInputArtifact(
                artifact_id=2,
                category=ArtifactCategory.SIGNING_INPUT,
                data=DebusineSigningInput(),
            ),
        )

    def test_get_provided_worker_tags(self) -> None:
        self.mock_is_command_available({"sbsign": True})
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "signing:sign:version:1",
                wtags.CAP_PREFIX + "sbsign",
            ],
        )

    def test_get_provided_worker_tags_sbsign_not_available(self) -> None:
        self.mock_is_command_available({"sbsign": False})
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [wtags.TASK_PREFIX + "signing:sign:version:1"],
        )

    def test_analyze_worker(self) -> None:
        """Test the analyze_worker() method."""
        self.mock_is_command_available({"sbsign": True})
        metadata = self.task.analyze_worker()
        self.assertEqual(metadata["signing:sign:available:uefi"], True)

    def test_analyze_worker_sbsign_not_available(self) -> None:
        """analyze_worker() handles sbsign not being available."""
        self.mock_is_command_available({"sbsign": False})
        metadata = self.task.analyze_worker()
        self.assertEqual(metadata["signing:sign:available:uefi"], False)

    def test_can_run_on_uefi(self) -> None:
        """can_run_on returns True for UEFI if sbsign is available."""
        self.assertTrue(
            self.task.can_run_on(
                {
                    "system:worker_type": WorkerType.SIGNING,
                    "signing:sign:available:uefi": True,
                    "signing:sign:version": self.task.TASK_VERSION,
                    "system:architectures": [FALLBACK_ARCHITECTURE],
                }
            )
        )

    def test_can_run_on_openpgp_upload(self) -> None:
        """can_run_on always returns True for OpenPGP upload keys."""
        self.task_data["purpose"] = KeyPurpose.OPENPGP_UPLOAD
        self.assertTrue(
            self.task.can_run_on(
                {
                    "system:worker_type": WorkerType.SIGNING,
                    "signing:sign:version": self.task.TASK_VERSION,
                    "system:architectures": [FALLBACK_ARCHITECTURE],
                }
            )
        )

    def test_can_run_on_mismatched_task_version(self) -> None:
        """can_run_on returns False for mismatched task versions."""
        self.assertFalse(
            self.task.can_run_on(
                {
                    "system:worker_type": WorkerType.SIGNING,
                    "signing:sign:available:uefi": True,
                    "signing:sign:version": self.task.TASK_VERSION + 1,
                }
            )
        )

    def test_can_run_on_uefi_missing_tool(self) -> None:
        """can_run_on returns False for UEFI if sbsign is not available."""
        self.assertFalse(
            self.task.can_run_on(
                {
                    "system:worker_type": WorkerType.SIGNING,
                    "signing:sign:available:uefi": False,
                    "signing:sign:version": self.task.TASK_VERSION,
                }
            )
        )

    def test_can_run_on_not_available(self) -> None:
        self.assertFalse(
            self.task.can_run_on(
                {
                    # "signing:sign:available:uefi" is not set
                    "system:worker_type": WorkerType.SIGNING,
                    "signing:sign:version": self.task.TASK_VERSION,
                    "system:architectures": [FALLBACK_ARCHITECTURE],
                }
            )
        )

    def test_compute_scheduler_tags_purpose_uefi(self) -> None:
        self.assertComputesSchedulerTags(
            require_system={wtags.CAP_PREFIX + "sbsign"}
        )

    def test_compute_scheduler_tags_purpose_openpgp(self) -> None:
        self.task_data["purpose"] = KeyPurpose.OPENPGP_UPLOAD
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            SignDynamicData(
                unsigned_ids=[2], unsigned_binary_package_names=[None]
            ),
        )

    def test_get_subject(self) -> None:
        self.assertIsNone(self.get_subject())

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_compute_dynamic_data_copies_binary_package_name(self) -> None:
        """Dynamic data copies `binary_package_name` from `unsigned`."""
        self.resolver.set_field_artifacts(
            "unsigned",
            BaseTaskInputArtifact(
                artifact_id=2,
                category=ArtifactCategory.SIGNING_INPUT,
                data=DebusineSigningInput(binary_package_name="hello"),
            ),
        )

        self.assertEqual(
            self.compute_dynamic_data(self.task),
            SignDynamicData(
                unsigned_ids=[2], unsigned_binary_package_names=["hello"]
            ),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = SignDynamicData(unsigned_ids=[2])
        self.assertEqual(self.task.get_input_artifacts_ids(), [2])

    def test_fetch_input_missing_trusted_cert(self) -> None:
        """fetch_input requires trusted_certs to match configuration."""
        trusted_cert = "0" * 64
        temp_directory = self.create_temporary_directory()
        (input_file := temp_directory / "data").touch()
        signing_input = SigningInputArtifact.create(
            [input_file], temp_directory, trusted_certs=[trusted_cert]
        )

        def download_artifact(
            artifact_id: int,
            destination: Path,  # noqa: ARG001, U100
            **kwargs: Any,
        ) -> ArtifactResponse:
            return self.make_artifact_response(signing_input, artifact_id)

        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": [2],
                "key": "ABC123",
            },
            dynamic_task_data={"unsigned_ids": [2]},
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.side_effect = download_artifact

        destination = self.create_temporary_directory()

        with override_settings(DEBUSINE_SIGNING_TRUSTED_CERTS=[]):
            self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"This installation cannot sign objects that trust certificate "
            f"'{trusted_cert}'\n",
        )

    def test_fetch_input_trusted_cert(self) -> None:
        """fetch_input accepts trusted_certs if they match configuration."""
        trusted_cert = "0" * 64
        fingerprint = "1" * 64
        key = Key.objects.create(
            purpose=Key.Purpose.UEFI,
            fingerprint=fingerprint,
            private_key={},
            public_key=b"",
        )
        temp_directory = self.create_temporary_directory()
        (input_file := temp_directory / "data").touch()
        signing_input = SigningInputArtifact.create(
            [input_file], temp_directory, trusted_certs=[trusted_cert]
        )
        signing_input_id = 2
        signing_input_response = self.make_artifact_response(
            signing_input, signing_input_id
        )

        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": [signing_input_id],
                "key": fingerprint,
            },
            dynamic_task_data={
                "unsigned_ids": [signing_input_id],
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = signing_input_response
        destination = self.create_temporary_directory()

        with override_settings(DEBUSINE_SIGNING_TRUSTED_CERTS=[trusted_cert]):
            self.assertTrue(task.fetch_input(destination))

        self.assertEqual(
            task._files, {signing_input_id: signing_input_response.files}
        )
        self.assertEqual(task._key, key)
        debusine_mock.download_artifact.assert_called_once_with(
            signing_input_id,
            destination,
            tarball=mock.ANY,
            workspace=task.workspace_name,
        )

    def test_fetch_input_mismatched_purpose(self) -> None:
        """fetch_input requires purpose to match key.purpose."""
        fingerprint = "0" * 64
        Key.objects.create(
            purpose=Key.Purpose.UEFI,
            fingerprint=fingerprint,
            private_key={},
        )

        temp_directory = self.create_temporary_directory()
        (input_file := temp_directory / "data").touch()
        signing_input = SigningInputArtifact.create(
            [input_file], temp_directory
        )
        signing_input_id = 2
        signing_input_response = self.make_artifact_response(
            signing_input, signing_input_id
        )

        task = Sign(
            task_data={
                "purpose": KeyPurpose.OPENPGP_UPLOAD,
                "unsigned": [signing_input_id],
                "key": fingerprint,
            },
            dynamic_task_data={
                "unsigned_ids": [signing_input_id],
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = signing_input_response
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"Signing key openpgp:{fingerprint} does not exist\n",
        )
        debusine_mock.download_artifact.assert_called_once_with(
            signing_input_id,
            destination,
            tarball=mock.ANY,
            workspace=task.workspace_name,
        )

    def test_fetch_input_missing_key(self) -> None:
        """fetch_input requires the signing key to exist in the database."""
        fingerprint = "0" * 64
        temp_directory = self.create_temporary_directory()
        (input_file := temp_directory / "data").touch()
        signing_input = SigningInputArtifact.create(
            [input_file], temp_directory
        )
        signing_input_id = 2
        signing_input_response = self.make_artifact_response(
            signing_input, signing_input_id
        )

        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": [signing_input_id],
                "key": fingerprint,
            },
            dynamic_task_data={
                "unsigned_ids": [signing_input_id],
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = signing_input_response
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"Signing key uefi:{fingerprint} does not exist\n",
        )
        debusine_mock.download_artifact.assert_called_once_with(
            signing_input_id,
            destination,
            tarball=mock.ANY,
            workspace=task.workspace_name,
        )

    def test_fetch_input_multiple_unsigned_ids(self) -> None:
        """fetch_input handles multiple unsigned artifacts."""
        fingerprint = "0" * 64
        key = Key.objects.create(
            purpose=Key.Purpose.UEFI,
            fingerprint=fingerprint,
            private_key={},
            public_key=b"",
        )
        signing_inputs = []
        for _ in range(2):
            temp_directory = self.create_temporary_directory()
            (input_file := temp_directory / "data").touch()
            signing_inputs.append(
                SigningInputArtifact.create([input_file], temp_directory)
            )
        signing_input_ids = [2, 3]
        responses = {
            artifact_id: self.make_artifact_response(artifact, artifact_id)
            for artifact, artifact_id in (
                (signing_inputs[0], signing_input_ids[0]),
                (signing_inputs[1], signing_input_ids[1]),
            )
        }

        def download_artifact(
            artifact_id: int,
            destination: Path,  # noqa: ARG001, U100
            **kwargs: Any,
        ) -> ArtifactResponse:
            return responses[artifact_id]

        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": signing_input_ids,
                "key": fingerprint,
            },
            dynamic_task_data={
                "unsigned_ids": signing_input_ids,
            },
        )
        task.work_request_id = 1
        task.workspace_name = "workspace"
        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.side_effect = download_artifact
        destination = self.create_temporary_directory()

        with override_settings(DEBUSINE_SIGNING_TRUSTED_CERTS=[]):
            self.assertTrue(task.fetch_input(destination))

        self.assertEqual(
            task._files,
            {
                signing_input_id: responses[signing_input_id].files
                for signing_input_id in signing_input_ids
            },
        )
        self.assertEqual(task._key, key)

    def test_fetch_input_permission_denied(self) -> None:
        fingerprint = "0" * 64
        Key.objects.create(
            purpose=Key.Purpose.UEFI,
            fingerprint=fingerprint,
            private_key={},
            public_key=b"",
        )
        temp_directory = self.create_temporary_directory()
        (input_file := temp_directory / "data").touch()
        signing_input = SigningInputArtifact.create(
            [input_file], temp_directory
        )
        signing_input_id = 2
        signing_input_response = self.make_artifact_response(
            signing_input, signing_input_id
        )

        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": [signing_input_id],
                "key": fingerprint,
            },
            dynamic_task_data={
                "unsigned_ids": [signing_input_id],
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = signing_input_response
        debusine_mock.asset_permission_check.return_value = (
            AssetPermissionCheckResponse(
                has_permission=False,
                username="testuser",
                user_id=123,
                resource={},
            )
        )
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"User testuser does not have permission to sign "
            f"{signing_input_id} with uefi:{fingerprint}.\n",
        )
        debusine_mock.asset_permission_check.assert_called_once_with(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=f"uefi:{fingerprint}",
            permission_name="sign_with",
            artifact_id=signing_input_id,
            work_request_id=task.work_request_id,
            workspace=task.workspace_name,
        )

    def test_run_error(self) -> None:
        """run() records an error result if signing fails."""
        key = Key.objects.create(
            purpose=Key.Purpose.UEFI,
            fingerprint="0" * 64,
            private_key={},
            public_key=b"",
        )
        download_directory = self.create_temporary_directory()
        execute_directory = self.create_temporary_directory()
        (image := download_directory / "image").touch()
        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": [2],
                "key": key.fingerprint,
            }
        )
        self.task_set_current(task)

        task._files = {2: self.make_files_response({"image": image})}
        task._permission_checks = {
            2: AssetPermissionCheckResponse(
                has_permission=True,
                username="testuser",
                user_id=123,
                resource={},
            )
        }
        task._key = key
        task.prepare_to_run(download_directory, execute_directory)

        with mock.patch(
            "debusine.signing.tasks.sign.sign", side_effect=Exception("Boom")
        ):
            self.assertEqual(
                task.run(execute_directory), WorkRequestResults.FAILURE
            )

        self.assertEqual(
            task._results,
            {2: [SigningResult(file="image", error_message="Boom")]},
        )

    def test_run_success(self) -> None:
        """run() records a successful result if signing succeeds."""
        key = Key.objects.create(
            purpose=Key.Purpose.UEFI,
            fingerprint="0" * 64,
            private_key={},
            public_key=b"",
        )
        download_directory = self.create_temporary_directory()
        execute_directory = self.create_temporary_directory()
        (image := download_directory / "image").touch()
        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": [2],
                "key": key.fingerprint,
            }
        )
        self.task_set_current(task)

        task._files = {2: self.make_files_response({"image": image})}
        task._permission_checks = {
            2: AssetPermissionCheckResponse(
                has_permission=True,
                username="testuser",
                user_id=123,
                resource={},
            )
        }
        task._key = key
        task.prepare_to_run(download_directory, execute_directory)

        with mock.patch("debusine.signing.tasks.sign.sign") as mock_sign:
            self.assertEqual(
                task.run(execute_directory), WorkRequestResults.SUCCESS
            )

        self.assertEqual(
            task._results,
            {2: [SigningResult(file="image", output_file="image.sig")]},
        )
        mock_sign.assert_called_once_with(
            [key],
            execute_directory / "input" / "image",
            execute_directory / "output" / "image.sig",
            SigningMode.DETACHED,
            task.work_request_id,
            username="testuser",
            user_id=123,
            resource={},
            log_file=mock.ANY,
        )

    def test_run_uefi_on_kmod(self) -> None:
        """run() skips kernel modules for UEFI signing."""
        key = Key.objects.create(
            purpose=Key.Purpose.UEFI,
            fingerprint="0" * 64,
            private_key={},
            public_key=b"",
        )
        download_directory = self.create_temporary_directory()
        execute_directory = self.create_temporary_directory()
        (image := download_directory / "image").touch()
        (foo_ko := download_directory / "foo.ko").touch()
        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": [2],
                "key": key.fingerprint,
            }
        )
        self.task_set_current(task)

        task._files = {
            2: self.make_files_response({"image": image, "foo.ko": foo_ko})
        }
        task._permission_checks = {
            2: AssetPermissionCheckResponse(
                has_permission=True,
                username="testuser",
                user_id=123,
                resource={},
            )
        }
        task._key = key
        task.prepare_to_run(download_directory, execute_directory)

        with mock.patch("debusine.signing.tasks.sign.sign"):
            self.assertEqual(
                task.run(execute_directory), WorkRequestResults.SUCCESS
            )

        self.assertEqual(
            task._results,
            {
                2: [
                    SigningResult(file="image", output_file="image.sig"),
                    SigningResult(
                        file="foo.ko",
                        error_message=(
                            "Skipped kernel module (not yet supported)"
                        ),
                    ),
                ]
            },
        )

    def test_run_uefi_on_only_kmod(self) -> None:
        """run() fails if no files are signed."""
        key = Key.objects.create(
            purpose=Key.Purpose.UEFI,
            fingerprint="0" * 64,
            private_key={},
            public_key=b"",
        )
        download_directory = self.create_temporary_directory()
        execute_directory = self.create_temporary_directory()
        (foo_ko := download_directory / "foo.ko").touch()
        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": [2],
                "key": key.fingerprint,
            }
        )
        self.task_set_current(task)

        task._files = {2: self.make_files_response({"foo.ko": foo_ko})}
        task._key = key
        task.prepare_to_run(download_directory, execute_directory)

        with mock.patch("debusine.signing.tasks.sign.sign"):
            self.assertEqual(
                task.run(execute_directory), WorkRequestResults.FAILURE
            )

        self.assertEqual(
            task._results,
            {
                2: [
                    SigningResult(
                        file="foo.ko",
                        error_message=(
                            "Skipped kernel module (not yet supported)"
                        ),
                    ),
                ]
            },
        )


class SignIntegrationTests(
    SigningTaskTestCase[Sign], TestCase, TransactionTestCase
):
    """Integration tests for the :py:class:`Sign` task."""

    @skipUnless(is_command_available("sbsign"), "requires sbsign")
    def test_uefi(self) -> None:
        """Integration test: sign using a real UEFI key."""
        service_private_key = PrivateKey.generate()
        with (
            override_settings(
                DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
            ),
            open(os.devnull, "wb") as log_file,
        ):
            key = Key.objects.generate(
                Key.Purpose.UEFI, "A UEFI key", 1, log_file
            )
        temp_path = self.create_temporary_directory()
        (minimal_pecoff_path := temp_path / "image").write_bytes(
            minimal_pecoff_image
        )
        (minimal_pecoff_path2 := temp_path / "image2").write_bytes(
            minimal_pecoff_image[:-1] + b"\0"
        )
        signing_inputs = [
            SigningInputArtifact.create([path], temp_path)
            for path in (minimal_pecoff_path, minimal_pecoff_path2)
        ]
        signing_input_ids = [2, 3]
        responses = {
            artifact_id: self.make_artifact_response(artifact, artifact_id)
            for artifact, artifact_id in (
                (signing_inputs[0], signing_input_ids[0]),
                (signing_inputs[1], signing_input_ids[1]),
            )
        }
        uploaded_paths: list[Path] = []
        output_path = self.create_temporary_directory()
        signing_output_ids = [4, 5]
        work_request_debug_logs_id = 6

        def download_artifact(
            artifact_id: int, destination: Path, **kwargs: Any
        ) -> ArtifactResponse:
            if artifact_id == signing_input_ids[0]:
                shutil.copy(minimal_pecoff_path, destination)
            elif artifact_id == signing_input_ids[1]:
                shutil.copy(minimal_pecoff_path2, destination)
            else:  # pragma: no cover
                pass
            return responses[artifact_id]

        def upload_artifact(
            local_artifact: LocalArtifact[Any], **kwargs: Any
        ) -> RemoteArtifact:
            for path in local_artifact.files.values():
                uploaded_paths.append(path)
                shutil.copy(path, output_path)
            match (local_artifact.category, list(local_artifact.files)[0]):
                case (ArtifactCategory.SIGNING_OUTPUT, "image.sig"):
                    artifact_id = signing_output_ids[0]
                case (ArtifactCategory.SIGNING_OUTPUT, "image2.sig"):
                    artifact_id = signing_output_ids[1]
                case (ArtifactCategory.WORK_REQUEST_DEBUG_LOGS, _):
                    artifact_id = work_request_debug_logs_id
                case _ as unreachable:
                    raise AssertionError(f"unexpected upload: {unreachable}")
            return create_remote_artifact(id=artifact_id, workspace="System")

        task = Sign(
            task_data={
                "purpose": KeyPurpose.UEFI,
                "unsigned": signing_input_ids,
                "key": key.fingerprint,
            },
            dynamic_task_data={
                "unsigned_ids": signing_input_ids,
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.side_effect = download_artifact
        debusine_mock.upload_artifact.side_effect = upload_artifact
        debusine_mock.asset_permission_check.return_value = (
            AssetPermissionCheckResponse(
                has_permission=True,
                username="testuser",
                user_id=123,
                resource={},
            )
        )

        with override_settings(
            DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
        ):
            self.assertEqual(task.execute(), WorkRequestResults.SUCCESS)

        # The task uploaded an artifact with the expected properties.
        self.assertEqual(len(uploaded_paths), 3)
        debusine_mock.upload_artifact.assert_has_calls(
            [
                mock.call(
                    SigningOutputArtifact(
                        category=SigningOutputArtifact._category,
                        files={"image.sig": uploaded_paths[0]},
                        content_types={"image.sig": "application/octet-stream"},
                        data=DebusineSigningOutput(
                            purpose=KeyPurpose.UEFI,
                            fingerprint=key.fingerprint,
                            results=[
                                SigningResult(
                                    file="image", output_file="image.sig"
                                )
                            ],
                        ),
                    ),
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
                mock.call(
                    SigningOutputArtifact(
                        category=SigningOutputArtifact._category,
                        files={"image2.sig": uploaded_paths[1]},
                        content_types={
                            "image2.sig": "application/octet-stream"
                        },
                        data=DebusineSigningOutput(
                            purpose=KeyPurpose.UEFI,
                            fingerprint=key.fingerprint,
                            results=[
                                SigningResult(
                                    file="image2", output_file="image2.sig"
                                )
                            ],
                        ),
                    ),
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
                mock.call(
                    WorkRequestDebugLogs(
                        category=WorkRequestDebugLogs._category,
                        data=EmptyArtifactData(),
                        files={
                            "cmd-output.log": (
                                task.log_directory / "cmd-output.log"
                            )
                        },
                        content_types={
                            "cmd-output.log": "text/plain; charset=utf-8"
                        },
                    ),
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
            ]
        )
        debusine_mock.relation_create.assert_has_calls(
            [
                mock.call(
                    signing_output_ids[0], signing_input_ids[0], "relates-to"
                ),
                mock.call(
                    signing_output_ids[1], signing_input_ids[1], "relates-to"
                ),
            ]
        )

        # The files were properly signed.
        (certificate := temp_path / "tmp.crt").write_bytes(key.public_key)
        for image_path, signature_path in (
            (minimal_pecoff_path, output_path / "image.sig"),
            (minimal_pecoff_path2, output_path / "image2.sig"),
        ):
            subprocess.run(
                [
                    "sbverify",
                    "--cert",
                    certificate,
                    "--detached",
                    signature_path,
                    image_path,
                ],
                check=True,
                stdout=subprocess.DEVNULL,
            )

    def test_openpgp_upload(self) -> None:
        """Integration test: sign using a real OpenPGP upload key."""
        service_private_key = PrivateKey.generate()
        with (
            override_settings(
                DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
            ),
            open(os.devnull, "wb") as log_file,
        ):
            key = Key.objects.generate(
                Key.Purpose.OPENPGP_UPLOAD, "An OpenPGP upload key", 1, log_file
            )
        temp_path = self.create_temporary_directory()
        (changes_path := temp_path / "foo.changes").write_text(
            "A changes file\n"
        )
        signing_input = SigningInputArtifact.create([changes_path], temp_path)
        signing_input_id = 2
        signing_input_response = self.make_artifact_response(
            signing_input, signing_input_id
        )
        uploaded_paths: list[Path] = []
        output_path = self.create_temporary_directory()
        signing_output_id = 3

        def download_artifact(
            artifact_id: int, destination: Path, **kwargs: Any
        ) -> ArtifactResponse:
            assert artifact_id == signing_input_id
            shutil.copy(changes_path, destination)
            return signing_input_response

        def upload_artifact(
            local_artifact: LocalArtifact[Any], **kwargs: Any
        ) -> RemoteArtifact:
            for path in local_artifact.files.values():
                uploaded_paths.append(path)
                shutil.copy(path, output_path)
            return create_remote_artifact(
                id=signing_output_id, workspace="System"
            )

        task = Sign(
            task_data={
                "purpose": KeyPurpose.OPENPGP_UPLOAD,
                "unsigned": [signing_input_id],
                "key": key.fingerprint,
            },
            dynamic_task_data={
                "unsigned_ids": [signing_input_id],
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.side_effect = download_artifact
        debusine_mock.upload_artifact.side_effect = upload_artifact
        debusine_mock.asset_permission_check.return_value = (
            AssetPermissionCheckResponse(
                has_permission=True,
                username="testuser",
                user_id=123,
                resource={},
            )
        )

        with override_settings(
            DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
        ):
            self.assertEqual(task.execute(), WorkRequestResults.SUCCESS)

        debusine_mock.asset_permission_check.assert_called_once_with(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=f"openpgp:{key.fingerprint}",
            permission_name="sign_with",
            artifact_id=signing_input_id,
            work_request_id=task.work_request_id,
            workspace=task.workspace_name,
        )

        # The task uploaded an artifact with the expected properties.
        self.assertEqual(len(uploaded_paths), 2)
        debusine_mock.upload_artifact.assert_has_calls(
            [
                mock.call(
                    SigningOutputArtifact(
                        category=SigningOutputArtifact._category,
                        files={"foo.changes.sig": uploaded_paths[0]},
                        content_types={
                            "foo.changes.sig": "application/octet-stream"
                        },
                        data=DebusineSigningOutput(
                            purpose=KeyPurpose.OPENPGP_UPLOAD,
                            fingerprint=key.fingerprint,
                            results=[
                                SigningResult(
                                    file="foo.changes",
                                    output_file="foo.changes.sig",
                                )
                            ],
                        ),
                    ),
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
                mock.call(
                    WorkRequestDebugLogs(
                        category=WorkRequestDebugLogs._category,
                        data=EmptyArtifactData(),
                        files={
                            "cmd-output.log": (
                                task.log_directory / "cmd-output.log"
                            )
                        },
                        content_types={
                            "cmd-output.log": "text/plain; charset=utf-8"
                        },
                    ),
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
            ]
        )
        debusine_mock.relation_create.assert_any_call(
            signing_output_id,
            signing_input_id,
            "relates-to",
        )

        # The file was properly signed.
        with (
            TemporaryDirectory(prefix="debusine-tests-") as tmp,
            gpg_ephemeral_context(Path(tmp)) as ctx,
        ):
            import_result = ctx.key_import(bytes(key.public_key))
            assert getattr(import_result, "imported", 0) == 1
            data, result = ctx.verify(
                (output_path / "foo.changes.sig").read_bytes()
            )
            self.assertEqual(data, b"A changes file\n")
            self.assertEqual(result.signatures[0].fpr, key.fingerprint)
            # We configure this in gpg_ephemeral_context, but make sure of
            # it.
            self.assertEqual(
                result.signatures[0].hash_algo, gpg.constants.MD_SHA512
            )



"""
Appropriate settings to run during development.

When running in development mode, selected.py should point to this file.
"""

from debusine.project.settings import defaults

try:
    from debusine.project.settings.db_postgresql import DATABASES
except (ModuleNotFoundError, PermissionError):
    from debusine.project.settings.db_postgresql_default import DATABASES

__all__ = [
    "ADMINS",
    "CACHES",
    "DATABASES",
    "DEBUG",
    "DEBUSINE_DEBIAN_ARCHIVE_FQDN",
    "DEBUSINE_DEBIAN_ARCHIVE_URL_SCHEME",
    "DEBUSINE_FQDN",
    "DEBUSINE_URL_SCHEME",
    "EMAIL_BACKEND",
    "INSTALLED_APPS",
    "INTERNAL_IPS",
    "MIDDLEWARE",
    "XHR_SIMULATED_DELAY",
]

DEBUG = True

ADMINS = ()

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'

DEBUSINE_FQDN = "127.0.0.1:8000"
DEBUSINE_URL_SCHEME = "http"
DEBUSINE_DEBIAN_ARCHIVE_FQDN = "127.0.0.1:8001"
DEBUSINE_DEBIAN_ARCHIVE_URL_SCHEME = "http"

CACHES = {
    'default': {'BACKEND': 'django.core.cache.backends.locmem.LocMemCache'}
}

XHR_SIMULATED_DELAY = 0.5

INSTALLED_APPS = defaults.INSTALLED_APPS.copy()
MIDDLEWARE = defaults.MIDDLEWARE.copy()
for optional_app, middleware in (
    ("debug_toolbar", "debug_toolbar.middleware.DebugToolbarMiddleware"),
    ("django_extensions", None),
):
    try:
        __import__(optional_app)
    except ImportError:
        pass
    else:
        INSTALLED_APPS.append(optional_app)
        if middleware:
            MIDDLEWARE.insert(0, middleware)

INTERNAL_IPS = ['127.0.0.1']



"""
Override path-related settings to use system wide paths.

The paths indicated in this file are those setup by the official Debian
package.
"""

__all__ = [
    'DEBUSINE_DATA_PATH',
]

DEBUSINE_DATA_PATH = '/var/lib/debusine/signing'



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data migration to create debian:archive in the default workspace."""

from django.conf import settings
from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps

from debusine.artifacts.models import CollectionCategory
from debusine.db.models import SYSTEM_USER_NAME
from debusine.db.models.collections import _CollectionItemTypes


def create_archive(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Create the debian:archive collection in the default workspace."""
    Workspace = apps.get_model("db", "Workspace")
    User = apps.get_model("db", "User")
    Collection = apps.get_model("db", "Collection")
    CollectionItem = apps.get_model("db", "CollectionItem")
    default_workspace = Workspace.objects.get(
        scope__name=settings.DEBUSINE_DEFAULT_SCOPE,
        name=settings.DEBUSINE_DEFAULT_WORKSPACE,
    )
    system_user = User.objects.get(username=SYSTEM_USER_NAME)
    archive = Collection.objects.create(
        name="_",
        category=CollectionCategory.ARCHIVE,
        workspace=default_workspace,
    )
    for suite in Collection.objects.filter(
        category=CollectionCategory.SUITE, workspace=default_workspace
    ):
        CollectionItem.objects.create(
            parent_collection=archive,
            parent_category=archive.category,
            name=suite.name,
            collection=suite,
            child_type=_CollectionItemTypes.COLLECTION,
            category=suite.category,
            data={},
            created_by_user=system_user,
        )


def remove_archive(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Remove the debian:archive collection in the default workspace."""
    Workspace = apps.get_model("db", "Workspace")
    Collection = apps.get_model("db", "Collection")
    CollectionItem = apps.get_model("db", "CollectionItem")
    default_workspace = Workspace.objects.get(
        scope__name=settings.DEBUSINE_DEFAULT_SCOPE,
        name=settings.DEBUSINE_DEFAULT_WORKSPACE,
    )
    archive_qs = Collection.objects.filter(
        name="_",
        category=CollectionCategory.ARCHIVE,
        workspace=default_workspace,
    )
    CollectionItem.objects.filter(parent_collection__in=archive_qs).delete()
    archive_qs.delete()


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0009_add_archive_singleton_collection"),
    ]
    replaces = [
        ("db", "0151_add_archive_singleton_collection_data"),
    ]

    operations = [migrations.RunPython(create_archive, remove_archive)]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Base structure of UI helpers for model instances."""

from collections.abc import Generator
from contextlib import contextmanager
from functools import cached_property
from typing import Any, Self, TYPE_CHECKING, cast, override

from django.db.models import Model
from django.http import HttpRequest

from debusine.db.context import context
from debusine.utils import extract_generic_type_arguments

if TYPE_CHECKING:
    from debusine.web.views.ui_shortcuts import UIShortcut

REQUEST_CACHE_NAME = "_debusine_ui_helpers"

#: Registry of helper classes indexed by model class
_helper_registry: dict[str, type["UI[Any]"]] = {}


def model_name(m: type[Model] | Model) -> str:
    """Return the model name for a model class or instance."""
    return m._meta.label_lower


class UI[M: Model]:
    """
    Per-model UI helpers.

    This is the root of a hierarchy of model companion classes that compute and
    cache per-request UI-specific information for a model instance.

    Helper classes are cached in the request and their lifetime corresponds to
    the request, so they can use @cached_property and cache other data in
    member variables at will.

    There is intentionally only one helper class per model, that acts as root
    for all use cases: this allows different UI needs to share intermediate
    cached results.
    """

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Register the subclass as the helper for M.

        The registry is used by the ``ui`` template tag.

        It is not used for instantiation from Python, which uses
        HelperClass.get() in order to give a correctly typed result.
        """
        (model_class,) = extract_generic_type_arguments(cls, UI)
        mname = model_name(model_class)
        if (old := _helper_registry.get(mname)) is not None:
            raise AssertionError(
                f"{old} already registered as helper for {model_class}"
            )
        _helper_registry[mname] = cls

    def __init__(self, request: HttpRequest, instance: M) -> None:
        """
        Instantiate the helper.

        :param request: the Django request
        :param ``instance``: the model instance
        """
        #: Current request
        self.request = request
        #: Model instance to wrap
        self.instance = instance

    def __getattr__(self, name: str) -> Any:
        """Proxy well-known model methods."""
        # Proxy and cache permission predicate checks
        if name.startswith("can_"):
            value = getattr(self.instance, name)(context.pc)
            setattr(self, name, value)
            return value
        # The attribute doesn't exist; let __getattribute__ raise
        # AssertionError with a suitable message.
        return self.__getattribute__(name)

    @classmethod
    def get(cls, request: HttpRequest, instance: M) -> Self:
        """
        Return the UI helper for a model instance.

        The result is cached in the request, and possibly reused.
        """
        # Get the cache dict from the request
        ui_helpers_cache: dict[tuple[str, int], UI[Any]] | None = getattr(
            request, REQUEST_CACHE_NAME, None
        )
        if ui_helpers_cache is None:
            ui_helpers_cache = {}
            setattr(request, REQUEST_CACHE_NAME, ui_helpers_cache)

        # Lookup an existing helper class in the cache
        cache_key = (model_name(instance), instance.pk)
        if cached := ui_helpers_cache.get(cache_key):
            return cast(Self, cached)

        # Cache miss: create it and cache it
        res = cls(request, instance)
        ui_helpers_cache[cache_key] = res
        return res

    @classmethod
    def for_instance(cls, request: HttpRequest, instance: M) -> "UI[M]":
        """
        Return the UI helper for a model instance.

        The result is cached in the request, and possibly reused.

        This is the factory used when the helper class is not known, such as in
        Django template tags. Using get() when possible means better typing of
        the result value.
        """
        helper_class = _helper_registry[model_name(instance)]
        return helper_class.get(request, instance)

    @classmethod
    @contextmanager
    def preserve_registry(cls) -> Generator[None]:
        """Undo changes to the helper registry on exit."""
        orig = _helper_registry.copy()
        try:
            yield
        finally:
            # Update in place, so that it also works for tests that imported
            # _helper_registry from base
            _helper_registry.clear()
            _helper_registry.update(orig)

    @cached_property
    def ui_shortcuts(self) -> list["UIShortcut"]:
        """Return UI shortcuts for this object."""
        return self.build_ui_shortcuts()

    def build_ui_shortcuts(self) -> list["UIShortcut"]:
        """Build the list of UI shortcuts for this object."""
        return []



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Default Django settings for the Debusine project.

Most settings are documented in this file and they are initialized to some
reasonable default values when possible.  They will be extended (and
possibly overridden) by settings from the other modules in this package
depending on the setup selected by the administrator. You likely won't
have to modify that file.

You should instead create local.py to put your site-specific settings (use
local.py.sample as template).

Here are the most important settings:

:py:data:`DEBUSINE_FQDN`
    The fully qualified domain name of the debusine installation.
    It should be a service-specific DNS entry like "debusine.example.com".
    Defaults to the FQDN of the machine which might not be adequate.

:py:data:`DEBUSINE_DATA_PATH`
    The directory where debusine will hold its data. The directory is
    further sub-divided in multiple directories for specific use
    cases (e.g. cache, keyring, static, media, logs, templates, etc.).
    Defaults to the "data" sub-directory in the debusine
    base directory (where the code lives).

:py:data:`MEDIA_URL`
    URL that handles the media served from MEDIA_ROOT. Make sure to use a
    trailing slash.
    Examples: "http://example.com/media/", "http://media.example.com/"
    Defaults to "/media/".

:py:data:`STATIC_URL`
    URL prefix for static files.
    Example: "http://example.com/static/", "http://static.example.com/"
    Defaults to "/static/"

Some settings have default values which are computed dynamically from
other settings. Those settings can also be overridden. Here's the list
of those settings.

py:data:`STATIC_ROOT`
    Absolute path to the directory static files should be collected to.
    Don't put anything in this directory yourself; store your static files
    in apps' "static/" subdirectories and in STATICFILES_DIRS. Defaults
    to the "static" sub-directory of :py:data:`DEBUSINE_DATA_PATH`.

:py:data:`MEDIA_ROOT`
    Absolute filesystem path to the directory that will hold user-uploaded
    files. Defaults to the "media" sub-directory of
    :py:data:`DEBUSINE_DATA_PATH`.

:py:data:`DEBUSINE_CACHE_DIRECTORY`
    This directory is used to store the locally cached resources.
    Any Debusine app should be able to use this directory to store
    its caches. For example, it is used to store the APT cache of repository
    information and the cache of retrieved Web resources.
    Defaults to the "cache" sub-directory of
    :py:data:`DEBUSINE_DATA_PATH`.

:py:data:`DEBUSINE_LOG_DIRECTORY`
    This directory will hold log files generated by debusine.
    Defaults to the "logs" sub-directory of py:data:`DEBUSINE_DATA_PATH`.

:py:data:`DEBUSINE_TEMPLATE_DIRECTORY`
    This directory can hold custom templates that will override the
    templates supplied by debusine. Defaults to the "templates"
    sub-directory of py:data:`DEBUSINE_DATA_PATH`.

:py:data:`DEBUSINE_TOKENS_DIRECTORY`
    This directory holds tokens for internal use by Debusine. Defaults to
    the "tokens" sub-directory of py:data:`DEBUSINE_DATA_PATH`.

:py:data:`DEBUSINE_UPLOAD_DIRECTORY`
    This directory temporarily holds files being uploaded to debusine.
    Defaults to the "uploads" sub-directory of py:data:`DEBUSINE_DATA_PATH`.

:py:data:`DEBUSINE_STORE_DIRECTORY`
    This directory is used by the default LocalFileBackend. Defaults to the
    "store" sub-directory of py:data:`DEBUSINE_DATA_PATH`.

:py:data:`DEBUSINE_FALLBACK_ARCHITECTURE`
    Architecture to use for tasks that require one but do not specify one. It
    defaults to ``amd64``.

More settings:

"""

import datetime as dt
import os
import os.path
import socket
from os.path import dirname
from typing import Any, TYPE_CHECKING

import django
from django.core.exceptions import ImproperlyConfigured

from debusine.django.django_utils import in_test_suite

if TYPE_CHECKING:
    from nacl.public import PrivateKey
else:
    PrivateKey = object

if django.VERSION < (4, 2):
    raise ImproperlyConfigured("Debusine needs Django >= 4.2")

# Django's debug mode, never enable this in production
DEBUG = False

BASE_DIR = dirname(dirname(dirname(dirname(__file__))))
DEBUSINE_DATA_PATH = os.path.join(BASE_DIR, 'data')

DEBUSINE_CACHE_DIRECTORY: str
DEBUSINE_TEMPLATE_DIRECTORY: str
DEBUSINE_LOG_DIRECTORY: str
DEBUSINE_UPLOAD_DIRECTORY: str
DEBUSINE_STORE_DIRECTORY: str
DEBUSINE_TOKENS_DIRECTORY: str

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# In a Windows environment this must be set to your system time zone.
TIME_ZONE = 'Etc/UTC'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://example.com/media/", "http://media.example.com/"
MEDIA_URL = '/media/'

# URL prefix for static files.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/static/'

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

STORAGES = {
    "default": {
        "BACKEND": "django.core.files.storage.FileSystemStorage",
    },
}

if in_test_suite():
    STORAGES["staticfiles"] = {
        "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage"
    }
else:
    STORAGES["staticfiles"] = {
        "BACKEND": "django.contrib.staticfiles.storage"
        ".ManifestStaticFilesStorage"
    }

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    # 'django.contrib.staticfiles.finders.DefaultStorageFinder',
]


STATICFILES_DIRS = [
    ("vendor/bootstrap5", "/usr/share/javascript/bootstrap5"),
    ("vendor/select2.js", "/usr/share/javascript/select2.js"),
    ("vendor/sortable", "/usr/share/javascript/sortable"),
    ("vendor/jquery", "/usr/share/javascript/jquery"),
]

# Overridden in production.py by content of /var/lib/debusine/server/key
SECRET_KEY = (
    "django-insecure-i77$9ld7x1%h1&_xpdjwy7i-jjaox@-ybtpe#g&8gw@yl4-1%&"
)

# Templating rules
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                # 'django.template.context_processors.i18n',
                # 'django.template.context_processors.media',
                'django.template.context_processors.static',
                # 'django.template.context_processors.tz',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'debusine.web.context_processors.server_info',
                'debusine.web.context_processors.application_context',
            ]
        },
    }
]

MIDDLEWARE: list[str] = [
    'django_prometheus.middleware.PrometheusBeforeMiddleware',
    'debusine.server.middlewares.virtual_hosts.VirtualHostMiddleware',
    # Introduces correct scoping for contextvars
    'debusine.server.middlewares.context.ContextMiddleware',
    'debusine.server.middlewares.headers.HeadersMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'debusine.server.middlewares.scopes.ScopeMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    # Needs to stay right after AuthenticationMiddleware
    'debusine.server.signon.middleware.SignonMiddleware',
    'debusine.server.middlewares.scopes.AuthorizationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    # Disabled to allow rendering in iframes
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django_prometheus.middleware.PrometheusAfterMiddleware',
]

AUTHENTICATION_BACKENDS = [
    'debusine.server.signon.auth.SignonAuthBackend',
]

LOGIN_REDIRECT_URL = "homepage:homepage"

# AUTHENTICATION_BACKENDS = [
#    'django_email_accounts.auth.UserEmailBackend',
# ]
#
# AUTH_USER_MODEL = 'accounts.User'

ROOT_URLCONF = 'debusine.project.urls'

# Python dotted path to the ASGI application used by Django's runserver.
ASGI_APPLICATION = "debusine.project.asgi.application"

# This is required to make request.is_secure() work behind a TLS-terminating
# proxy.
SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')

INSTALLED_APPS: list[str] = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_celery_results',
    'django_prometheus',
    'channels',
    'pgtrigger',
    'rest_framework',
    'debusine.db',
    'debusine.server',
    'debusine.web',
]

# This will need to be tightened up later.
PAGE_CLASS = 'DEFAULT_PAGINATION_CLASS'
REST_FRAMEWORK = {
    PAGE_CLASS: 'rest_framework.pagination.PageNumberPagination',
    'PAGE_SIZE': 10,
    "EXCEPTION_HANDLER": "debusine.server.exceptions."
    "debusine_exception_handler",
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "debusine.server.views.auth.DebusineTokenAuthentication",
        "rest_framework.authentication.SessionAuthentication",
    ],
}

# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': (
            'django.contrib.auth.password_validation.'
            'UserAttributeSimilarityValidator'
        )
    },
    {
        'NAME': (
            'django.contrib.auth.password_validation.MinimumLengthValidator'
        )
    },
    {
        'NAME': (
            'django.contrib.auth.password_validation.CommonPasswordValidator'
        )
    },
    {
        'NAME': (
            'django.contrib.auth.password_validation.NumericPasswordValidator'
        )
    },
]

# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s [%(module)s/%(process)d/%(thread)d] '
            + '%(levelname)s: %(message)s'
        },
        'standard': {
            'format': '%(asctime)s %(process)d %(levelname)s: %(message)s'
        },
        'simple': {'format': '%(asctime)s %(levelname)s: %(message)s'},
        'access': {
            '()': 'uvicorn.logging.AccessFormatter',
            'fmt': '%(client_addr)s - "%(request_line)s" %(status_code)s',
        },
    },
    'filters': {
        'require_debug_false': {'()': 'django.utils.log.RequireDebugFalse'},
        'require_debug_true': {'()': 'django.utils.log.RequireDebugTrue'},
        'suppress_db_errors': {
            '()': 'debusine.django.log_filters.SuppressDatabaseErrors'
        },
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
            'filters': ['require_debug_true'],
        },
        'access': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'access',
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false', 'suppress_db_errors'],
            'class': 'django.utils.log.AdminEmailHandler',
        },
        'null': {'level': 'DEBUG', 'class': 'logging.NullHandler'},
        'mail.log': {
            'level': 'INFO',
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'mail.log',
            'encoding': 'utf-8',
            'formatter': 'standard',
            'when': 'W0',
            'backupCount': 52,
        },
        'tasks.log': {
            'level': 'INFO',
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'tasks.log',
            'encoding': 'utf-8',
            'formatter': 'standard',
            'when': 'W0',
            'backupCount': 52,
        },
        'errors.log': {
            'level': 'WARNING',
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'errors.log',
            'encoding': 'utf-8',
            'formatter': 'verbose',
            'when': 'W0',
            'backupCount': 52,
        },
        'debug.log': {
            'level': 'DEBUG',
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'debug.log',
            'encoding': 'utf-8',
            'formatter': 'verbose',
            'when': 'W0',
            'backupCount': 52,
        },
    },
    'loggers': {
        'root': {
            'handlers': ['errors.log'],
            'level': 'ERROR',
            'propagate': False,
        },
        'uvicorn': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': False,
        },
        'uvicorn.access': {
            'handlers': ['access'],
            'level': 'INFO',
            'propagate': False,
        },
        'django.request': {
            'handlers': ['errors.log', 'mail_admins'],
            'level': 'ERROR',
            'propagate': False,
        },
        'django.security': {
            'handlers': ['errors.log', 'mail_admins'],
            'level': 'ERROR',
            'propagate': False,
        },
        'django.security.DisallowedHost': {
            'handlers': ['errors.log'],
            'level': 'ERROR',
            'propagate': False,
        },
        'py.warnings': {'handlers': ['console']},
        'debusine': {
            'handlers': ['debug.log', 'errors.log', 'console'],
            'level': 'DEBUG',
            'propagate': False,
        },
        'debusine.mail': {
            'handlers': ['mail.log', 'mail_admins'],
            'level': 'DEBUG',
            'propagate': True,
        },
        'debusine.tasks': {
            'handlers': ['tasks.log', 'mail_admins'],
            'level': 'DEBUG',
            'propagate': True,
        },
    },
}

AUTH_USER_MODEL = "db.User"

CELERY_BEAT_SCHEDULE = {
    # Run the scheduler once every fifteen seconds, just in case we have a
    # bug where something forgot to request a run of the scheduler when data
    # changed, or in case a scheduler run hit a lock and couldn't make
    # progress.
    "fallback-scheduler": {
        "task": "debusine.server.scheduler.schedule_task",
        "schedule": dt.timedelta(seconds=15),
    },
    "provisioner": {
        "task": "debusine.server.provisioning.provision",
        "schedule": dt.timedelta(minutes=1),
    },
    "worker-pool-scrub": {
        "task": "debusine.server.provisioning.scrub",
        "schedule": dt.timedelta(minutes=15),
    },
}
CELERY_BROKER_URL = "redis://localhost:6379/0"
CELERY_RESULT_BACKEND = "django-db"
CELERY_RESULT_BACKEND_TRANSPORT_OPTIONS = {"global_keyprefix": "celery"}
CELERY_RESULT_EXTENDED = True
CELERY_TASK_ROUTES = {
    "debusine.server.scheduler.schedule_task": {"queue": "scheduler"},
    "debusine.server.provisioning.provision": {"queue": "provisioner"},
    "debusine.server.provisioning.scrub": {"queue": "provisioner"},
}
CELERY_WORKER_PREFETCH_MULTIPLIER = 1
# Control memory leaks.
CELERY_WORKER_MAX_MEMORY_PER_CHILD = 512 * 1024  # in KiB, so 512 MiB

PROMETHEUS_REDIS_URL = "redis://localhost:6379/1"

EMAIL_TIMEOUT = 5

PGTRANSACTION_RETRY = 3

# https://github.com/django-commons/django-prometheus/issues/281
PROMETHEUS_EXPORT_MIGRATIONS = False

# *_created Prometheus metrics are useless to us.  Disable them.
os.environ.setdefault("PROMETHEUS_DISABLE_CREATED_SERIES", "1")

# === Debusine specific settings ===

# The fully qualified domain name for the Debusine deployment
DEBUSINE_FQDN: str = socket.getfqdn()

# URL scheme to use to construct Debusine URLs
DEBUSINE_URL_SCHEME: str = "https"

# The fully qualified domain name for Debian-format archives hosted by this
# deployment
DEBUSINE_DEBIAN_ARCHIVE_FQDN: str | list[str] = f"deb.{DEBUSINE_FQDN}"

# URL scheme to use to construct debian archive URLs
DEBUSINE_DEBIAN_ARCHIVE_URL_SCHEME: str = "https"

# Default CHANNEL_LAYERS pointing at a Redis local configuration
CHANNEL_LAYERS: dict[str, dict[str, Any]] = {
    "default": {
        "BACKEND": "channels_redis.pubsub.RedisPubSubChannelLayer",
        "CONFIG": {"hosts": ["redis://localhost:6379"]},
    }
}

# A default contact email
DEBUSINE_CONTACT_EMAIL = f'owner@{DEBUSINE_FQDN}'

# Name of the default scope to use for legacy web requests and API calls that
# are unaware of scopes
DEBUSINE_DEFAULT_SCOPE = 'debusine'
# Name of the default workspace to use for migrations, API calls that do not
# specify the workspace, management commands, and so on.
DEBUSINE_DEFAULT_WORKSPACE = "System"

# Encryption keys used by the signing service to encrypt private keys; the
# first in the list is used for newly-encrypted private keys, while the
# others may also be used for decryption
DEBUSINE_SIGNING_PRIVATE_KEYS: list[PrivateKey] = []

# Allow signing objects that declare a trust chain to these certificate
# fingerprints.  See
# https://wiki.debian.org/SecureBoot/Discussion#Describing_the_trust_chain.
DEBUSINE_SIGNING_TRUSTED_CERTS: list[str] = []

# Architecture to use for tasks that require one but do not specify one.
DEBUSINE_FALLBACK_ARCHITECTURE: str = "amd64"

# The lambda functions are evaluated at the end of the settings import
# logic. They provide default values to settings which have not yet been
# set (neither above nor in local.py).
_COMPUTE_DEFAULT_SETTINGS = (
    (
        'ALLOWED_HOSTS',
        lambda t: (
            [t['DEBUSINE_FQDN']]
            + (
                t['DEBUSINE_DEBIAN_ARCHIVE_FQDN']
                if isinstance(t['DEBUSINE_DEBIAN_ARCHIVE_FQDN'], list)
                else [t['DEBUSINE_DEBIAN_ARCHIVE_FQDN']]
            )
            + [
                'localhost',
                '127.0.0.1',
            ]
        ),
    ),
    (
        'DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN',
        lambda t: (
            t['DEBUSINE_DEBIAN_ARCHIVE_FQDN'][0]
            if isinstance(t['DEBUSINE_DEBIAN_ARCHIVE_FQDN'], list)
            else t['DEBUSINE_DEBIAN_ARCHIVE_FQDN']
        ),
    ),
    ('ADMINS', lambda t: (('Debusine Admins', t['DEBUSINE_CONTACT_EMAIL']),)),
    ('SERVER_EMAIL', lambda t: t['DEBUSINE_CONTACT_EMAIL']),
    ('DEFAULT_FROM_EMAIL', lambda t: t['DEBUSINE_CONTACT_EMAIL']),
    ('STATIC_ROOT', lambda t: os.path.join(t['DEBUSINE_DATA_PATH'], 'static')),
    ('MEDIA_ROOT', lambda t: os.path.join(t['DEBUSINE_DATA_PATH'], 'media')),
    (
        'DEBUSINE_CACHE_DIRECTORY',
        lambda t: os.path.join(t['DEBUSINE_DATA_PATH'], 'cache'),
    ),
    (
        'DEBUSINE_TEMPLATE_DIRECTORY',
        lambda t: os.path.join(t['DEBUSINE_DATA_PATH'], 'templates'),
    ),
    (
        'DEBUSINE_LOG_DIRECTORY',
        lambda t: os.path.join(t['DEBUSINE_DATA_PATH'], 'logs'),
    ),
    (
        'DEBUSINE_UPLOAD_DIRECTORY',
        lambda t: os.path.join(t['DEBUSINE_DATA_PATH'], 'uploads'),
    ),
    (
        'DEBUSINE_STORE_DIRECTORY',
        lambda t: os.path.join(t['DEBUSINE_DATA_PATH'], 'store'),
    ),
    (
        'DEBUSINE_TOKENS_DIRECTORY',
        lambda t: os.path.join(t['DEBUSINE_DATA_PATH'], 'tokens'),
    ),
)


def compute_default_settings(target: dict[str, Any]) -> None:
    """
    Dynamically generate some default settings.

    There are many settings whose default value depends on another
    setting. They are defined in the _COMPUTE_DEFAULT_SETTINGS dict
    with a function that evaluates them.
    """
    for setting, value in _COMPUTE_DEFAULT_SETTINGS:
        if setting in target:
            continue  # Settings is already defined
        target[setting] = value(target)
    # Extend TEMPLATE_DIRS with our directory
    target['TEMPLATES'][0]['DIRS'].append(target['DEBUSINE_TEMPLATE_DIRECTORY'])
    # Update LOGGING with full paths
    for handler in target['LOGGING']['handlers'].values():
        if 'filename' not in handler or "/" in handler['filename']:
            continue
        handler['filename'] = os.path.join(
            target['DEBUSINE_LOG_DIRECTORY'], handler['filename']
        )
    # Update DATABASES with full paths
    dbconf = target['DATABASES']['default']
    if dbconf['ENGINE'] == 'django.db.backends.sqlite3':
        if '/' not in dbconf['NAME']:
            dbconf['NAME'] = os.path.join(
                target['DEBUSINE_DATA_PATH'], dbconf['NAME']
            )
        if (
            'TEST' in dbconf
            and 'NAME' in dbconf['TEST']
            and '/' not in dbconf['TEST']['NAME']
        ):
            dbconf['TEST']['NAME'] = os.path.join(
                target['DEBUSINE_DATA_PATH'], dbconf['TEST']['NAME']
            )


LOGGING_CONFIG = "debusine.django.django_utils.test_data_override"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine signing tasks.

The tasks in this package run in a signing worker.

Signing tasks are executed with a dedicated user whose home directory is
locked to external users with restrictive permissions (chmod 0700). Due to
this, signing tasks can't use most container-based executors that
typically require processes using sub-uids to enter the working directory
(which by default is stored below other the signing worker's home
directory).
"""

from debusine.signing.tasks.base import BaseSigningTask  # isort: split

# Sub-tasks need to be imported in order to be available to BaseExternalTask
# (e.g. for BaseExternalTask.is_valid_task_name). They are registered via
# BaseExternalTask.__init_subclass__.
from debusine.signing.tasks.debsign import Debsign
from debusine.signing.tasks.generate_key import GenerateKey
from debusine.signing.tasks.noop import SigningNoop
from debusine.signing.tasks.sign import Sign
from debusine.signing.tasks.sign_repository_index import SignRepositoryIndex

__all__ = [
    "BaseSigningTask",
    "Debsign",
    "GenerateKey",
    "Sign",
    "SignRepositoryIndex",
    "SigningNoop",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command create_work_request."""

import io
from typing import Any

import yaml
from django.conf import settings
from django.core.management import CommandError

from debusine.artifacts.models import TaskTypes
from debusine.db.models import (
    User,
    WorkRequest,
    Workspace,
    default_workspace,
    system_user,
)
from debusine.django.management.tests import call_command
from debusine.tasks.models import ActionSendNotification, EventReactions
from debusine.test.django import TestCase


class CreateWorkRequestCommandTests(TestCase):
    """Tests for the create_work_request command."""

    def assert_work_request(
        self,
        work_request: WorkRequest,
        *,
        workspace: Workspace | None = None,
        created_by: User | None = None,
        task_data: dict[str, Any] | None = None,
        event_reactions: EventReactions | None = None,
    ) -> None:
        self.assertEqual(
            work_request.workspace, workspace or default_workspace()
        )
        self.assertEqual(work_request.created_by, created_by or system_user())
        self.assertEqual(work_request.status, WorkRequest.Statuses.PENDING)
        self.assertEqual(work_request.task_data, task_data or {})
        self.assertEqual(
            work_request.event_reactions, event_reactions or EventReactions()
        )

    def test_data_from_file(self) -> None:
        """`create_work_request` accepts data from a file."""
        data = {"result": WorkRequest.Results.SUCCESS}
        data_file = self.create_temporary_file(
            contents=yaml.safe_dump(data).encode()
        )
        stdout, stderr, exit_code = call_command(
            "create_work_request",
            "worker",
            "noop",
            "--data",
            str(data_file),
        )

        work_request = WorkRequest.objects.get(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.assert_work_request(work_request, task_data=data)

        expected = yaml.safe_dump({"work_request_id": work_request.id})
        self.assertEqual(stdout, f"---\n{expected}\n")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_data_from_stdin(self) -> None:
        """`create_work_request` accepts data from stdin."""
        data = {"result": WorkRequest.Results.SUCCESS}
        stdout, stderr, exit_code = call_command(
            "create_work_request",
            "worker",
            "noop",
            stdin=io.StringIO(yaml.safe_dump(data)),
        )

        work_request = WorkRequest.objects.get(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.assert_work_request(work_request, task_data=data)

        expected = yaml.safe_dump({"work_request_id": work_request.id})
        self.assertEqual(stdout, f"---\n{expected}\n")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_event_reactions(self) -> None:
        """`create_work_request` handles event reactions."""
        data = {"result": WorkRequest.Results.SUCCESS}
        event_reactions = {
            "on_success": [{"action": "send-notification", "channel": "foo"}]
        }
        event_reactions_file = self.create_temporary_file(
            contents=yaml.safe_dump(event_reactions).encode()
        )
        stdout, stderr, exit_code = call_command(
            "create_work_request",
            "worker",
            "noop",
            "--event-reactions",
            str(event_reactions_file),
            stdin=io.StringIO(yaml.safe_dump(data)),
        )

        work_request = WorkRequest.objects.get(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.assert_work_request(
            work_request,
            task_data=data,
            event_reactions=EventReactions(
                on_success=[ActionSendNotification(channel="foo")]
            ),
        )

        expected = yaml.safe_dump({"work_request_id": work_request.id})
        self.assertEqual(stdout, f"---\n{expected}\n")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_server(self) -> None:
        data = {"result": WorkRequest.Results.SUCCESS}
        stdout, stderr, exit_code = call_command(
            "create_work_request",
            "server",
            "servernoop",
            stdin=io.StringIO(yaml.safe_dump(data)),
        )

        work_request = WorkRequest.objects.get(
            task_type=TaskTypes.SERVER, task_name="servernoop"
        )
        self.assert_work_request(work_request, task_data=data)

        expected = yaml.safe_dump({"work_request_id": work_request.id})
        self.assertEqual(stdout, f"---\n{expected}\n")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_server_no_event_reactions(self) -> None:
        data = {"result": WorkRequest.Results.SUCCESS}
        event_reactions = {
            "on_success": [
                {
                    "action": "update-collection-with-artifacts",
                    "collection": 123,
                    "artifact_filters": {"category": "test"},
                }
            ]
        }
        event_reactions_file = self.create_temporary_file(
            contents=yaml.safe_dump(event_reactions).encode()
        )
        with self.assertRaisesRegex(
            CommandError, r"^Event reactions are not supported for SERVER tasks"
        ) as exc:
            call_command(
                "create_work_request",
                "server",
                "servernoop",
                "--event-reactions",
                str(event_reactions_file),
                stdin=io.StringIO(yaml.safe_dump(data)),
            )
        self.assertEqual(exc.exception.returncode, 1)

    def test_signing(self) -> None:
        """`create_work_request` can create signing work requests."""
        data = {"purpose": "uefi", "description": "A UEFI key"}
        stdout, stderr, exit_code = call_command(
            "create_work_request",
            "signing",
            "generatekey",
            stdin=io.StringIO(yaml.safe_dump(data)),
        )

        work_request = WorkRequest.objects.get(
            task_type=TaskTypes.SIGNING, task_name="generatekey"
        )
        self.assert_work_request(work_request, task_data=data)

        expected = yaml.safe_dump({"work_request_id": work_request.id})
        self.assertEqual(stdout, f"---\n{expected}\n")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_signing_no_event_reactions(self) -> None:
        data = {"purpose": "uefi", "description": "A UEFI key"}
        event_reactions = {
            "on_success": [
                {
                    "action": "update-collection-with-artifacts",
                    "collection": 123,
                    "artifact_filters": {"category": "test"},
                }
            ]
        }
        event_reactions_file = self.create_temporary_file(
            contents=yaml.safe_dump(event_reactions).encode()
        )
        with self.assertRaisesRegex(
            CommandError,
            r"^Event reactions are not supported for SIGNING tasks",
        ) as exc:
            call_command(
                "create_work_request",
                "signing",
                "generatekey",
                "--event-reactions",
                str(event_reactions_file),
                stdin=io.StringIO(yaml.safe_dump(data)),
            )
        self.assertEqual(exc.exception.returncode, 1)

    def test_wait_not_implemented(self) -> None:
        """`create_work_request` refuses to create wait work requests."""
        data = {"result": WorkRequest.Results.SUCCESS}
        with self.assertRaisesRegex(
            CommandError, r"^Error: argument task_type: invalid choice: 'wait'"
        ) as exc:
            call_command(
                "create_work_request",
                "wait",
                "externaldebsign",
                stdin=io.StringIO(yaml.safe_dump(data)),
            )
        self.assertEqual(exc.exception.returncode, 1)

        self.assertQuerySetEqual(
            WorkRequest.objects.filter(
                task_type=TaskTypes.WAIT, task_name="externaldebsign"
            ),
            [],
        )

    def test_empty_data(self) -> None:
        """`create_work_request` defaults data to {}."""
        stdout, stderr, exit_code = call_command(
            "create_work_request", "worker", "noop", stdin=io.StringIO()
        )

        work_request = WorkRequest.objects.get(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.assert_work_request(work_request, task_data={})

        expected = yaml.safe_dump({"work_request_id": work_request.id})
        self.assertEqual(stdout, f"---\n{expected}\n")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_different_created_by(self) -> None:
        """`create_work_request` can use a different created-by user."""
        user = self.playground.get_default_user()
        data = {"result": WorkRequest.Results.SUCCESS}
        stdout, stderr, exit_code = call_command(
            "create_work_request",
            "worker",
            "noop",
            "--created-by",
            user.username,
            stdin=io.StringIO(yaml.safe_dump(data)),
        )

        work_request = WorkRequest.objects.get(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.assert_work_request(work_request, created_by=user, task_data=data)

        expected = yaml.safe_dump({"work_request_id": work_request.id})
        self.assertEqual(stdout, f"---\n{expected}\n")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_different_workspace(self) -> None:
        """`create_work_request` can use a non-default workspace."""
        workspace_name = "test-workspace"
        workspace = self.playground.create_workspace(name=workspace_name)
        stdout, stderr, exit_code = call_command(
            "create_work_request",
            "worker",
            "noop",
            "--workspace",
            workspace_name,
            stdin=io.StringIO(),
        )

        work_request = WorkRequest.objects.get(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.assert_work_request(work_request, workspace=workspace)

        expected = yaml.safe_dump({"work_request_id": work_request.id})
        self.assertEqual(stdout, f"---\n{expected}\n")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_different_scope(self) -> None:
        """`create_work_request` can use a workspace in a different scope."""
        scope_name = "test-scope"
        scope = self.playground.get_or_create_scope(name=scope_name)
        workspace = self.playground.create_workspace(
            scope=scope, name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        stdout, stderr, exit_code = call_command(
            "create_work_request",
            "worker",
            "noop",
            "--workspace",
            f"{scope_name}/{settings.DEBUSINE_DEFAULT_WORKSPACE}",
            stdin=io.StringIO(),
        )

        work_request = WorkRequest.objects.get(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.assert_work_request(work_request, workspace=workspace)

        expected = yaml.safe_dump({"work_request_id": work_request.id})
        self.assertEqual(stdout, f"---\n{expected}\n")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_invalid_data_yaml(self) -> None:
        """`create_work_request` returns error: cannot parse data."""
        with self.assertRaisesRegex(
            CommandError, r"^Error parsing YAML:"
        ) as exc:
            call_command(
                "create_work_request", "worker", "noop", stdin=io.StringIO(":")
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_invalid_event_reactions_yaml(self) -> None:
        """`create_work_request` returns error: cannot parse data."""
        event_reactions_file = self.create_temporary_file(contents=b":")

        with self.assertRaisesRegex(
            CommandError, r"^Error parsing YAML:"
        ) as exc:
            call_command(
                "create_work_request",
                "worker",
                "noop",
                "--event-reactions",
                str(event_reactions_file),
                stdin=io.StringIO(),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_bad_task_type(self) -> None:
        """`create_work_request` returns error: bad task type."""
        with self.assertRaisesRegex(
            CommandError, r"task_type: invalid choice: 'nonexistent'"
        ) as exc:
            call_command(
                "create_work_request",
                "nonexistent",
                "noop",
                stdin=io.StringIO(),
            )

        self.assertEqual(exc.exception.returncode, 1)

    def test_user_not_found(self) -> None:
        """`create_work_request` returns error: user not found."""
        with self.assertRaisesRegex(
            CommandError, r'^User "nonexistent" not found'
        ) as exc:
            call_command(
                "create_work_request",
                "worker",
                "noop",
                "--created-by",
                "nonexistent",
                stdin=io.StringIO(),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_workspace_not_found(self) -> None:
        """`create_work_request` returns error: workspace not found."""
        with self.assertRaisesRegex(
            CommandError, r"^Workspace 'nonexistent' not found"
        ) as exc:
            call_command(
                "create_work_request",
                "worker",
                "noop",
                "--workspace",
                "nonexistent",
                stdin=io.StringIO(),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_task_name_not_found(self) -> None:
        """`create_work_request` returns error: task name not found."""
        with self.assertRaisesRegex(
            CommandError,
            r"'task_name': \['nonexistent: invalid Worker task name'\]",
        ) as exc:
            call_command(
                "create_work_request",
                "worker",
                "nonexistent",
                stdin=io.StringIO(),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_bad_task_data(self) -> None:
        """`create_work_request` returns error: bad task data."""
        with self.assertRaisesRegex(
            CommandError, r"invalid Worker task data"
        ) as exc:
            call_command(
                "create_work_request",
                "worker",
                "noop",
                stdin=io.StringIO("foo: bar\n"),
            )

        self.assertEqual(exc.exception.returncode, 3)
        self.assertFalse(
            WorkRequest.objects.filter(
                task_type=TaskTypes.WORKER, task_name="noop"
            ).exists()
        )
